{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "class VM(object):\n",
    "    def __init__(self, register, ip=-1):\n",
    "        self.ip = ip\n",
    "        self.ip_binding = None\n",
    "        self.register = register\n",
    "        self.instructions = []\n",
    "        self.opcodes = {\n",
    "            \"addr\": self.addr,\n",
    "            \"addi\": self.addi,\n",
    "            \"mulr\": self.mulr,\n",
    "            \"muli\": self.muli,\n",
    "            \"banr\": self.banr,\n",
    "            \"bani\": self.bani,\n",
    "            \"borr\": self.borr,\n",
    "            \"bori\": self.bori,\n",
    "            \"setr\": self.setr,\n",
    "            \"seti\": self.seti,\n",
    "            \"gtir\": self.gtir,\n",
    "            \"gtri\": self.gtri,\n",
    "            \"gtrr\": self.gtrr,\n",
    "            \"eqir\": self.eqir,\n",
    "            \"eqri\": self.eqri,\n",
    "            \"eqrr\": self.eqrr\n",
    "        }\n",
    "\n",
    "    def load(self, instructions):\n",
    "        self.instructions = []\n",
    "        for instruction in instructions:\n",
    "            if \"#ip\" == instruction[:3]:\n",
    "                _, ip_str = instruction.split(\" \")\n",
    "                self.ip_binding = int(ip_str)\n",
    "            else:\n",
    "                op, a, b, c = instruction.split(\" \")\n",
    "                self.instructions.append([op, int(a), int(b), int(c)])\n",
    "\n",
    "    def get_debug_state(self):\n",
    "        return \" \".join([\n",
    "            f\"ip={self.ip}\",\n",
    "            str(self.register),\n",
    "            \" \".join(str(i) for i in self.instructions[self.ip])\n",
    "        ])\n",
    "        \n",
    "    def run(self, debug=False):\n",
    "        while True:\n",
    "            try:\n",
    "                self.ip += 1\n",
    "                debug_state = self.get_debug_state()\n",
    "                self.step()\n",
    "                if debug:\n",
    "                    print(debug_state, str(self.register))            \n",
    "            except:\n",
    "                break\n",
    "                \n",
    "        return self.register\n",
    "        \n",
    "    def step(self):       \n",
    "        if self.ip >= len(self.instructions):\n",
    "            return False\n",
    "        \n",
    "        op, a, b, c = self.instructions[self.ip]\n",
    "        self.execute(op, a, b, c)\n",
    "            \n",
    "        return True\n",
    "        \n",
    "    def execute(self, opcode, a, b, c):\n",
    "        if self.ip_binding != None:\n",
    "            self.register[self.ip_binding] = self.ip\n",
    "            \n",
    "        decode_int = {\n",
    "            0 : self.eqir,\n",
    "            1 : self.seti,\n",
    "            2 : self.eqri,\n",
    "            3 : self.eqrr,\n",
    "            4 : self.addi,\n",
    "            5 : self.setr,\n",
    "            6 : self.gtrr,\n",
    "            7 : self.gtri,\n",
    "            8 : self.muli,\n",
    "            9 : self.bori,\n",
    "            10: self.bani,\n",
    "            11: self.borr,\n",
    "            12: self.gtir,\n",
    "            13: self.banr,\n",
    "            14: self.addr,\n",
    "            15: self.mulr\n",
    "        }\n",
    "        \n",
    "        decode_str = {\n",
    "            \"eqir\" : self.eqir,\n",
    "            \"seti\" : self.seti,\n",
    "            \"eqri\" : self.eqri,\n",
    "            \"eqrr\" : self.eqrr,\n",
    "            \"addi\" : self.addi,\n",
    "            \"setr\" : self.setr,\n",
    "            \"gtrr\" : self.gtrr,\n",
    "            \"gtri\" : self.gtri,\n",
    "            \"muli\" : self.muli,\n",
    "            \"bori\" : self.bori,\n",
    "            \"bani\" : self.bani,\n",
    "            \"borr\" : self.borr,\n",
    "            \"gtir\" : self.gtir,\n",
    "            \"banr\" : self.banr,\n",
    "            \"addr\" : self.addr,\n",
    "            \"mulr\" : self.mulr\n",
    "        }\n",
    "        \n",
    "        if (type(opcode) is str):\n",
    "            assert(decode_str[opcode](a,b,c))\n",
    "        elif (type(opcode) is int):\n",
    "            assert(decode_int[opcode](a,b,c))\n",
    "        else:\n",
    "            raise Exception(f\"opcode {opcode} is neither int nor str\")\n",
    "        \n",
    "        if self.ip_binding != None:\n",
    "            self.ip = self.register[self.ip_binding]\n",
    "        \n",
    "    def any_invalid_registers(self, regs):\n",
    "        result = True\n",
    "\n",
    "        for reg in regs:\n",
    "            result = result & (reg < len(self.register))\n",
    "\n",
    "        return not result\n",
    "\n",
    "    def test_opcode(self, opcode_name, initial_register_state, a, b, c, final_register_state):\n",
    "        self.register = [ reg for reg in initial_register_state ]\n",
    "        result = self.opcodes[opcode_name](a, b, c)\n",
    "        return result and (self.register == final_register_state)\n",
    "\n",
    "    # addr (add register) stores into register C the result of adding register A and register B.\n",
    "    def addr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a,b,c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] + self.register[b]\n",
    "        return True\n",
    "\n",
    "    # addi (add immediate) stores into register C the result of adding register A and value B.\n",
    "    def addi(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False    \n",
    "\n",
    "        self.register[c] = self.register[a] + b\n",
    "        return True\n",
    "\n",
    "    # mulr (multiply register) stores into register C the result of multiplying register A and register B.\n",
    "    def mulr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] * self.register[b]\n",
    "        return True\n",
    "\n",
    "    # muli (multiply immediate) stores into register C the result of multiplying register A and value B.\n",
    "    def muli(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] * b\n",
    "        return True\n",
    "\n",
    "    # banr (bitwise AND register) stores into register C the result of the bitwise AND of register A and register B.\n",
    "    def banr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] & self.register[b]\n",
    "        return True\n",
    "\n",
    "    # bani (bitwise AND immediate) stores into register C the result of the bitwise AND of register A and value B.\n",
    "    def bani(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] & b\n",
    "        return True\n",
    "\n",
    "    # borr (bitwise OR register) stores into register C the result of the bitwise OR of register A and register B.\n",
    "    def borr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] | self.register[b]\n",
    "        return True\n",
    "\n",
    "\n",
    "    # bori (bitwise OR immediate) stores into register C the result of the bitwise OR of register A and value B.\n",
    "    def bori(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] | b\n",
    "        return True\n",
    "\n",
    "    # setr (set register) copies the contents of register A into register C. (Input B is ignored.)\n",
    "    def setr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a]\n",
    "        return True\n",
    "\n",
    "    # seti (set immediate) stores value A into register C. (Input B is ignored.)\n",
    "    def seti(self, a, b, c):\n",
    "        if self.any_invalid_registers([c]):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = a\n",
    "        return True\n",
    "\n",
    "    # gtir (greater-than immediate/register) sets register C to 1 if value A is greater than register B. Otherwise, register C is set to 0.\n",
    "    def gtir(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(a > self.register[b])\n",
    "        return True\n",
    "\n",
    "    # gtri (greater-than register/immediate) sets register C to 1 if register A is greater than value B. Otherwise, register C is set to 0.\n",
    "    def gtri(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, a)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] > b)\n",
    "        return True\n",
    "\n",
    "    # gtrr (greater-than register/register) sets register C to 1 if register A is greater than register B. Otherwise, register C is set to 0.\n",
    "    def gtrr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] > self.register[b])\n",
    "        return True\n",
    "\n",
    "    # eqir (equal immediate/register) sets register C to 1 if value A is equal to register B. Otherwise, register C is set to 0.\n",
    "    def eqir(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(a == self.register[b])\n",
    "        return True\n",
    "\n",
    "    # eqri (equal register/immediate) sets register C to 1 if register A is equal to value B. Otherwise, register C is set to 0.\n",
    "    def eqri(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, a)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] == b)\n",
    "        return True\n",
    "\n",
    "    # eqrr (equal register/register) sets register C to 1 if register A is equal to register B. Otherwise, register C is set to 0.\n",
    "    def eqrr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a,b,c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] == self.register[b])\n",
    "        return True\n",
    "    \n",
    "    def disassemble(self):\n",
    "        return \"\\n\".join(\n",
    "            f\"{i:02}: \" + self.disassemble_instr(instr)\n",
    "            for i, instr in enumerate(self.instructions)\n",
    "        )\n",
    "    \n",
    "    def disassemble_instr(self, instr):\n",
    "        op, a_, b_, c = instr\n",
    "        \n",
    "        a = self.disassemble_reg(a_, True)\n",
    "        ra = self.disassemble_reg(a_, False)\n",
    "        b = self.disassemble_reg(b_, True)\n",
    "        rb = self.disassemble_reg(b_, False)\n",
    "        rc = self.disassemble_reg(c, False)\n",
    "        \n",
    "        diss_instr = {\n",
    "            \"addr\": \"{ra} + {rb}\",\n",
    "            \"addi\": \"{ra} + {b}\",\n",
    "            \"mulr\": \"{ra} * {rb}\",\n",
    "            \"muli\": \"{ra} * {b}\",\n",
    "            \"banr\": \"{ra} & {rb}\",\n",
    "            \"bani\": \"{ra} & {b}\",\n",
    "            \"borr\": \"{ra} | {rb}\",\n",
    "            \"bori\": \"{ra} | {b}\",\n",
    "            \"setr\": \"{ra}\",\n",
    "            \"seti\": \"{a}\",\n",
    "            \"gtrr\": \"{ra} > {rb} ? 1 : 0\",\n",
    "            \"gtri\": \"{ra} > {b} ? 1 : 0\",\n",
    "            \"eqir\": \"{a} == {rb} ? 1 : 0\",\n",
    "            \"eqri\": \"{ra} == {b} ? 1 : 0\",    \n",
    "            \"eqrr\": \"{ra} == {rb} ? 1 : 0\",    \n",
    "        }\n",
    "        \n",
    "        return f\"{rc} = \" + diss_instr[op].format(a=a, ra=ra, b=b, rb=rb)\n",
    "    \n",
    "    def disassemble_reg(self, reg, immediate):\n",
    "        if self.ip_binding != None and reg == self.ip_binding:\n",
    "            return \"IP\"\n",
    "        elif immediate:\n",
    "            return str(reg)\n",
    "        return f\"R{reg}\""
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "scrolled": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1806\n"
     ]
    }
   ],
   "source": [
    "puzzle_input_str = open(\"puzzle_input/day19.txt\").read()\n",
    "test_input_str = \"\"\"#ip 0\n",
    "seti 5 0 1\n",
    "seti 6 0 2\n",
    "addi 0 1 0\n",
    "addr 1 2 3\n",
    "setr 1 0 0\n",
    "seti 8 0 4\n",
    "seti 9 0 5\"\"\"\n",
    "\n",
    "def part_one(raw_instructions, registers, debug=False):\n",
    "    instructions = raw_instructions.split(\"\\n\")\n",
    "    vm = VM(registers)\n",
    "    vm.load(instructions)\n",
    "    return vm.run(debug)[0]\n",
    "\n",
    "assert 6 == part_one(test_input_str, 6 * [0])\n",
    "    \n",
    "# part 1\n",
    "print(part_one(puzzle_input_str, [0] * 6))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "00: IP = IP + 16\n",
      "01: R1 = 1\n",
      "02: R4 = 1\n",
      "03: R3 = R1 * R4\n",
      "04: R3 = R3 == R2 ? 1 : 0\n",
      "05: IP = R3 + IP\n",
      "06: IP = IP + 1\n",
      "07: R0 = R1 + R0\n",
      "08: R4 = R4 + 1\n",
      "09: R3 = R4 > R2 ? 1 : 0\n",
      "10: IP = IP + R3\n",
      "11: IP = 2\n",
      "12: R1 = R1 + 1\n",
      "13: R3 = R1 > R2 ? 1 : 0\n",
      "14: IP = R3 + IP\n",
      "15: IP = 1\n",
      "16: IP = IP * IP\n",
      "17: R2 = R2 + 2\n",
      "18: R2 = R2 * R2\n",
      "19: R2 = IP * R2\n",
      "20: R2 = R2 * 11\n",
      "21: R3 = R3 + 8\n",
      "22: R3 = R3 * IP\n",
      "23: R3 = R3 + 16\n",
      "24: R2 = R2 + R3\n",
      "25: IP = IP + R0\n",
      "26: IP = 0\n",
      "27: R3 = IP\n",
      "28: R3 = R3 * IP\n",
      "29: R3 = IP + R3\n",
      "30: R3 = IP * R3\n",
      "31: R3 = R3 * 14\n",
      "32: R3 = R3 * IP\n",
      "33: R2 = R2 + R3\n",
      "34: R0 = 0\n",
      "35: IP = 0\n"
     ]
    }
   ],
   "source": [
    "# part 2\n",
    "# print(run_program(problem_input_str, [1] + 5 * [0], False))\n",
    "# FFS ok so this is basically a super-quadratic program that actually requires you to\n",
    "# disassemble your code, identify what the program is _trying_ to do and find a better\n",
    "# approach. That's frustrating.\n",
    "def disassemble_program(raw_instructions, registers, debug=False):\n",
    "    instructions = raw_instructions.split(\"\\n\")\n",
    "    vm = VM(registers)\n",
    "    vm.load(instructions)\n",
    "    print(vm.disassemble())\n",
    "    \n",
    "disassemble_program(puzzle_input_str, [0] * 6)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "18741072"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# reassembled into python this looks like the below\n",
    "def naive_reassembly():\n",
    "    r0, r1, r2, r3, r4, r5 = [1, 0, 0, 0, 0, 0]\n",
    "\n",
    "    # initialise r2 to 10551428 very inefficiently\n",
    "    r2 = r2 + 2\n",
    "    r2 = r2 * r2\n",
    "    r2 = 19 * r2\n",
    "    r2 = r2 * 11\n",
    "    r3 = r3 + 8\n",
    "    r3 = r3 * 22\n",
    "    r3 = r3 + 16\n",
    "    r2 = r2 + r3\n",
    "    r3 = 27\n",
    "    r3 = r3 * 28\n",
    "    r3 = 29 + r3\n",
    "    r3 = 30 * r3\n",
    "    r3 = r3 * 14\n",
    "    r3 = r3 * 32\n",
    "    r2 = r2 + r3\n",
    "\n",
    "    r1 = 1\n",
    "    while r1 <= r2:\n",
    "        r4 = 1    \n",
    "        while r4 <= r2:\n",
    "            if r1 * r4 == r2:\n",
    "                r0 += r1\n",
    "            r4 += 1\n",
    "        r1 += 1\n",
    "        \n",
    "    return r0\n",
    "\n",
    "# ... but this is still slow as hell. after a bit of fiddling around it became \n",
    "# apparent we're basically just summing up all the integers which are factors\n",
    "# of 10551428. this is a very low \n",
    "def part_two():\n",
    "    total = 0\n",
    "    n = 10551428\n",
    "\n",
    "    while n > 0:\n",
    "        if 10551428 % n == 0:\n",
    "            total += n\n",
    "        n -= 1\n",
    "\n",
    "    return total\n",
    "\n",
    "part_two()"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
